module Main
    where

import List
import Char
import Monad
import Maybe
import System

{-
STPP, the array preprocessor.  The syntax is simple:

we want to write:

  y <- 3 * a[|x|] + 1
becomes
  y <- readArray a x >>= return . (+1) . (3*)

so

  y <- e a[|x|] f
becomes
  y <- readArray a x >>= return . (f) . (e)

similarly, we want to be able to do assignment, so

  a[|x|] <- exp
becomes
  writeArray a x exp

and for monadic operations:

  a[|x|] <<- exp
becomes
  exp >>= writeArray a x

finally, we want to do updates:

  a[|x|] <-$ exp
becomes
  readArray a x >>= exp >>= writeArray a x

we also want to be able to handle "2d" arrays, so, for assignment:

  a[|x|][|y|] <- exp
becomes
  exp >>= \v -> readArray a x >>= \arr -> writeArray arr y v

or, in general:

  a[|x1|]...[|xn|] <- exp
becomes
  exp >>= \v -> readArray a x1 >>= (flip readArray) x2 >>= 
                (flip readArray) x3   >>= ...
                (flip readArray) xn-1 >>= 
                \arr -> writeArray arr xn v

for update, it's easy:

  a[|x1|]...[|xn|] <-$ exp
becomes
  readArray a x1 >>= (flip readArray) x2   >>= 
                     (flip readArray) x3   >>= ...
                     (flip readArray) xn-1 >>= 
                     \arr -> readArray arr xn >>=
		     exp >>= writeArray arr xn

and finally, for a read:

  y <- e a[|x1|][|x2|]...[|xn|] f
becomes
  y <- readArray a x1 >>=
       (flip readArray) x2 >>= ...
       (flip readArray) xn >>= return . (f) . (e)


----------------------------------------------------------------------
----------------------------------------------------------------------
----------------------------------------------------------------------

we also support hashing operations:

  y <- e h{|x1|}{|x2|}...{|xn|} f
becomes
  y <- lookupHT a x1 >>=
       (flip maybeLookupHT) x2 >>= ...
       (flip maybeLookupHT) xn >>= return . (f) . (e)

and so on for the other operations
-}

trim = reverse . trim' . reverse . trim'
    where trim' [] = []
	  trim' (x:xs) | isSpace x = trim' xs
		       | otherwise = x:xs

doMonadicAssignment :: String -> String
doMonadicAssignment l = 
    case xl of
      [] -> prews ++ exp ++ " >>= writeArray " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ exp ++ " >>= \\stpp_value -> readArray " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip readArray) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_arr -> writeArray stpp_arr (" ++ last (xl) ++ ") stpp_value"
    where (pre1,exp') = splitUntil "<<-" l
	  exp          = trim $ drop 3 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "[|" al
	  x1:xl     = extractXL xl' []

doAssignment :: String -> String
doAssignment l = 
    case xl of
      [] -> prews ++ "return " ++ exp ++ " >>= writeArray " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ "return " ++ exp ++ " >>= \\stpp_value -> readArray " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip readArray) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_arr -> writeArray stpp_arr (" ++ last (xl) ++ ") stpp_value"
    where (pre1,exp') = splitUntil "<-" l
	  exp          = trim $ drop 2 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "[|" al
	  x1:xl     = extractXL xl' []

doUpdate :: String -> String
doUpdate l = 
    case xl of
      [] -> prews ++ "readArray " ++ a ++ " (" ++ x1 ++ ") >>= " ++ exp ++ " >>= writeArray " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ "readArray " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip readArray) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_arr -> readArray stpp_arr (" ++ last xl ++ ") >>= " ++ exp ++ " >>= " ++
		   "writeArray stpp_arr (" ++ last xl ++ ")"
    where (pre1,exp') = splitUntil "<-$" l
	  exp          = trim $ drop 3 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "[|" al
	  x1:xl     = extractXL xl' []


doRead :: String -> String
doRead l = prews ++ y ++ " <<- readArray " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		     unwords (map (\xi -> "(flip readArray) (" ++ xi ++ ") >>= ") xl) ++
		     "return" ++ f' ++ e' ++ postline
    where f' = case f of
	         [] -> []
		 x  -> " . (" ++ x ++ ")"
          e' = case e of
	         [] -> []
		 x  -> " . (" ++ x ++ ")"
          (pre1,post1) = splitUntil "<<-" l
	  y = trim $ last $ words pre1
	  prews = fst $ splitUntil y pre1
	  (pre,post)= splitUntil "[|" (drop 3 post1)
	  a         = trim $ reverse $ takeWhile (not . isSpace) $ reverse pre
	  (xl',f2)  = 
	      let barInd = elemIndices '|' post
		  clInd  = elemIndices ']' post
		  maxV   = maximum (map (+1) barInd `intersect` clInd)
	      in  splitAt (maxV+1) post
	  f = dropToEnd (0,0,0) f2
	  postline = drop (length f) f2
	  x1:xl     = extractXL xl' []
	  e         = trim $ fst $ splitUntil a pre


dropToEnd (np,nb,nr) [] = []
dropToEnd (-1,_,_) _ = []
dropToEnd (_,-1,_) _ = []
dropToEnd (_,_,-1) _ = []
dropToEnd (np,nb,nr) ('(':xs) = '(' : dropToEnd (np+1,nb,nr) xs
dropToEnd (np,nb,nr) (')':xs) = ')' : dropToEnd (np-1,nb,nr) xs
dropToEnd (np,nb,nr) ('[':xs) = '[' : dropToEnd (np,nb+1,nr) xs
dropToEnd (np,nb,nr) (']':xs) = ']' : dropToEnd (np,nb-1,nr) xs
dropToEnd (np,nb,nr) ('{':xs) = '{' : dropToEnd (np,nb,nr+1) xs
dropToEnd (np,nb,nr) ('}':xs) = '}' : dropToEnd (np,nb,nr-1) xs
dropToEnd c (x:xs) = x : dropToEnd c xs

doUseVal l = pre ++ "readArray (" ++ a ++ ") (" ++ x1 ++ ") " ++
	       unwords (map (\xi -> ">>= (flip readArray) (" ++ xi ++ ") ") xl) ++
	       rest
    where (pre1,post) = splitUntil "[|" l
	  a = trim $ reverse $ takeWhile (not . isSpace) $ reverse pre1
	  pre = fst $ splitUntil a pre1
	  (xl',rest) = splitAt (sLastIndexOf "|]" post+2) post
	  x1:xl = extractXL xl' []

doHashMonadicAssignment :: String -> String
doHashMonadicAssignment l = 
    case xl of
      [] -> prews ++ exp ++ " >>= addHT " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ exp ++ " >>= \\stpp_value -> lookupHT " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip maybeLookupHT) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_hash -> addHT stpp_hash (" ++ last (xl) ++ ") stpp_value"
    where (pre1,exp') = splitUntil "<<-" l
	  exp          = trim $ drop 3 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "{|" al
	  x1:xl     = extractXLHash xl' []

doHashAssignment :: String -> String
doHashAssignment l = 
    case xl of
      [] -> prews ++ "return " ++ exp ++ " >>= addHT " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ "return " ++ exp ++ " >>= \\stpp_value -> lookupHT " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip maybeLookupHT) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_hash -> addHT stpp_hash (" ++ last (xl) ++ ") stpp_value"
    where (pre1,exp') = splitUntil "<-" l
	  exp          = trim $ drop 2 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "{|" al
	  x1:xl     = extractXLHash xl' []

doHashUpdate :: String -> String
doHashUpdate l = 
    case xl of
      [] -> prews ++ "lookupHT " ++ a ++ " (" ++ x1 ++ ") >>= " ++ exp ++ " >>= addHT " ++ a ++ " (" ++ x1 ++ ")"
      _  -> prews ++ "lookupHT " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		   unwords (map (\xi -> "(flip maybeLookupHT) (" ++ xi ++ ") >>= ") (allButLast xl)) ++
		   "\\stpp_hash -> maybeLookupHT stpp_hash (" ++ last xl ++ ") >>= " ++ exp ++ " >>= " ++
		   "addHT stpp_hash (" ++ last xl ++ ")"
    where (pre1,exp') = splitUntil "<-$" l
	  exp          = trim $ drop 3 exp'
	  al           = reverse $ takeWhile (not . isSpace) $ dropWhile isSpace $ reverse pre1
	  prews = fst $ splitUntil al pre1
	  (a,xl') = splitUntil "{|" al
	  x1:xl     = extractXLHash xl' []


doHashRead :: String -> String
doHashRead l = prews ++ y ++ " <<- lookupHT " ++ a ++ " (" ++ x1 ++ ") >>= " ++
		     unwords (map (\xi -> "(flip maybeLookupHT) (" ++ xi ++ ") >>= ") xl) ++
		     "return" ++ f' ++ e' ++ postline
    where f' = case f of
	         [] -> []
		 x  -> " . (" ++ x ++ ")"
          e' = case e of
	         [] -> []
		 x  -> " . (" ++ x ++ ")"
          (pre1,post1) = splitUntil "<<-" l
	  y = trim $ last $ words pre1
	  prews = fst $ splitUntil y pre1
	  (pre,post)= splitUntil "{|" (drop 3 post1)
	  a         = trim $ reverse $ takeWhile (not . isSpace) $ reverse pre
	  (xl',f2)  = 
	      let barInd = elemIndices '|' post
		  clInd  = elemIndices '}' post
		  maxV   = maximum (map (+1) barInd `intersect` clInd)
	      in  splitAt (maxV+1) post
	  f = dropToEnd (0,0,0) f2
	  postline = drop (length f) f2
	  x1:xl     = extractXLHash xl' []
	  e         = trim $ fst $ splitUntil a pre

	  
doHashUseVal l = pre ++ "lookupHT (" ++ a ++ ") (" ++ x1 ++ ") " ++
	       unwords (map (\xi -> ">>= (flip maybeLookupHT) (" ++ xi ++ ") ") xl) ++
	       rest
    where (pre1,post) = splitUntil "{|" l
	  a = trim $ reverse $ takeWhile (not . isSpace) $ reverse pre1
	  pre = fst $ splitUntil a pre1
	  (xl',rest) = splitAt (sLastIndexOf "|}" post+2) post
	  x1:xl = extractXLHash xl' []
	  

doArrayLine l
    | "<-$" `isSubstringOf` l = doLine (doUpdate l)
    | "<<-"  `isSubstringOf` l &&
      sIndexOf "<<-" l < sLastIndexOf "[|"l = doLine (doRead l)
    | "<<-" `isSubstringOf` l = doLine (doMonadicAssignment l)
    | "<-"  `isSubstringOf` l = doLine (doAssignment l)
    | otherwise               = doLine (doUseVal l)

doHashLine l
    | "<-$" `isSubstringOf` l = doLine (doHashUpdate l)
    | "<<-" `isSubstringOf` l &&
      sIndexOf "<<-" l < sLastIndexOf "{|" l = doLine (doHashRead l)
    | "<<-" `isSubstringOf` l = doLine (doHashMonadicAssignment l)
    | "<-"  `isSubstringOf` l = doLine (doHashAssignment l)
    | otherwise               = doLine (doHashUseVal l)

doMif :: String -> String
doMif l = pre1 ++ takeWhile isSpace w ++ trim exp' ++ " >>= \\stpp_mif_val -> if stpp_mif_val" ++ rest'
    where (pre1, w , exp)  = splitUntilW "mif" l
          (exp', w2, rest) = splitUntilW "then" exp
          rest'            = w2 ++ rest
          
doMcase :: String -> String
doMcase l = pre1 ++ takeWhile isSpace w ++ trim exp' ++ " >>= \\stpp_mcase_val -> case stpp_mcase_val" ++ rest'
    where (pre1, w , exp)  = splitUntilW "mcase" l
          (exp', w2, rest) = splitUntilW "of" exp
          rest'            = w2 ++ rest

doLine :: String -> String
doLine l
    | "[|"  `isSubstringOf` l &&
      "|]"  `isSubstringOf` l = doArrayLine l
    | "{|"  `isSubstringOf` l &&
      "|}"  `isSubstringOf` l = doHashLine l
    | "mif" `isWSubstringOf` l = doMif l
    | "mcase" `isWSubstringOf` l = doMcase l
    | otherwise               = l

sLastIndexOf sub = sIndexOf' 0 (-1)
    where sIndexOf' _ acc [] = acc
          sIndexOf' ind acc l | sub `isPrefixOf` l = sIndexOf' (ind+1) ind (tail l)
			      | otherwise = sIndexOf' (ind+1) acc (tail l)

sIndexOf sub = sIndexOf' 0
    where sIndexOf' _ [] = -1
          sIndexOf' ind l | sub `isPrefixOf` l = ind
			  | otherwise = sIndexOf' (ind+1) (tail l)

short `isSubstringOf` long = sIndexOf short long >= 0
short `isWSubstringOf` long = 
    let i = sIndexOf short long
        h = long !! (i-1)
        f = long !! (i + length short)
    in  (i == 0 ||
         (i >= 1 && isSeparator h)) &&
        (i == length long - length short ||
         isSeparator f)

allButLast [] = []
allButLast [x] = []
allButLast (x:xs) = x : allButLast xs

extractXL [] _ = []
extractXL ('[':'|':xs) _   = extractXL xs []
extractXL ('|':']':xs) acc = reverse acc : extractXL xs []
extractXL (x:xs) acc       = extractXL xs (x:acc)

extractXLHash [] _ = []
extractXLHash ('{':'|':xs) _   = extractXLHash xs []
extractXLHash ('|':'}':xs) acc = reverse acc : extractXLHash xs []
extractXLHash (x:xs) acc       = extractXLHash xs (x:acc)

splitUntil s [] = ([],[])
splitUntil s z@(x:xs)
    | s `isPrefixOf` z = ([],z)
    | otherwise        = let (pre,post) = splitUntil s xs
			 in  (x:pre,post)

splitUntilW s z
    | s `isPrefixOf` z,
      rest <- drop (length s) z,
      (null rest || isSeparator (head rest)) = ([], s ++ take 1 rest, drop 1 rest)
    | otherwise = splitUntilW' s z

splitUntilW' s [] = ([],[],[])
splitUntilW' s z@(x:xs)
    | isSeparator x,
      s `isPrefixOf` xs,
      rest <- drop (length s) xs,
      (null rest || isSeparator (head rest)) = ([], x : s ++ take 1 rest, drop 1 rest)
    | otherwise        = let (pre,w,post) = splitUntilW' s xs
			 in  (x:pre,w,post)

isSeparator x = not (isAlphaNum x) && not (x `elem` "'?")

main =
    do args <- getArgs
       case args of
        (_:inp:outp:_) -> readFile inp >>= writeFile outp . unlines . map doLine . lines
	_ -> interact (unlines . map doLine . lines)
